# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT
from __future__ import absolute_import
import collections

import pytest
import solv

from module_build_service.web.mmd_resolver import MMDResolver
from tests import make_module


class TestMMDResolver:
    def setup_method(self, test_method):
        self.mmd_resolver = MMDResolver()

    def teardown_method(self, test_method):
        pass

    @pytest.mark.parametrize(
        "deps, expected",
        (
            ([], "None"),
            ([{"x": []}], "module(x)"),
            ([{"x": ["1"]}], "(module(x) with module(x:1))"),
            ([{"x": ["1", "2"]}], "(module(x) with (module(x:1) or module(x:2)))"),
            ([{"x": [], "y": []}], "(module(x) and module(y))"),
            ([{"x": []}, {"y": []}], "(module(x) or module(y))"),
        ),
    )
    def test_deps2reqs(self, deps, expected):
        # Sort by keys here to avoid unordered dicts
        deps = [collections.OrderedDict(sorted(dep.items())) for dep in deps]
        reqs = self.mmd_resolver._deps2reqs(deps)
        assert str(reqs) == expected

    @pytest.mark.parametrize(
        "dependencies, expected",
        (
            (
                [{"buildrequires": {"platform": []}}],
                [
                    [
                        ["platform:f28:0:c0:x86_64"],
                        ["platform:f29:0:c0:x86_64"],
                    ]
                ]
            ),
            (
                [{"buildrequires": {"platform": ["f28"]}}],
                [
                    [["platform:f28:0:c0:x86_64"]]
                ]
            ),
            (
                [{"buildrequires": {"gtk": [], "qt": []}}],
                [
                    [
                        ["gtk:3:0:c8:x86_64", "qt:4:0:c8:x86_64", "platform:f28:0:c0:x86_64"],
                        ["gtk:4:0:c8:x86_64", "qt:4:0:c8:x86_64", "platform:f28:0:c0:x86_64"],
                        ["gtk:3:0:c8:x86_64", "qt:5:0:c8:x86_64", "platform:f28:0:c0:x86_64"],
                        ["gtk:4:0:c8:x86_64", "qt:5:0:c8:x86_64", "platform:f28:0:c0:x86_64"],
                    ]
                ],
            ),
            (
                [{"buildrequires": {"gtk": [], "qt": [], "platform": []}}],
                [
                    [
                        ["gtk:3:0:c8:x86_64", "qt:4:0:c8:x86_64", "platform:f28:0:c0:x86_64"],
                        ["gtk:4:0:c8:x86_64", "qt:4:0:c8:x86_64", "platform:f28:0:c0:x86_64"],
                        ["gtk:3:0:c8:x86_64", "qt:5:0:c8:x86_64", "platform:f28:0:c0:x86_64"],
                        ["gtk:4:0:c8:x86_64", "qt:5:0:c8:x86_64", "platform:f28:0:c0:x86_64"],
                        ["gtk:3:0:c9:x86_64", "qt:4:0:c9:x86_64", "platform:f29:0:c0:x86_64"],
                        ["gtk:4:0:c9:x86_64", "qt:4:0:c9:x86_64", "platform:f29:0:c0:x86_64"],
                        ["gtk:3:0:c9:x86_64", "qt:5:0:c9:x86_64", "platform:f29:0:c0:x86_64"],
                        ["gtk:4:0:c9:x86_64", "qt:5:0:c9:x86_64", "platform:f29:0:c0:x86_64"],
                    ]
                ],
            ),
            (
                [
                    {"buildrequires": {"qt": [], "platform": ["f28"]}},
                    {"buildrequires": {"gtk": [], "platform": ["f29"]}}
                ],
                [
                    [
                        ["qt:4:0:c8:x86_64", "platform:f28:0:c0:x86_64"],
                        ["qt:5:0:c8:x86_64", "platform:f28:0:c0:x86_64"],
                    ],
                    [
                        ["gtk:3:0:c9:x86_64", "platform:f29:0:c0:x86_64"],
                        ["gtk:4:0:c9:x86_64", "platform:f29:0:c0:x86_64"],
                    ],
                ],
            ),
            (
                [{"buildrequires": {"mess": []}}],
                [[["mess:1:0:c0:x86_64", "gtk:3:0:c8:x86_64", "platform:f28:0:c0:x86_64"]]],
            ),
            (
                [{"buildrequires": {"mess": [], "platform": []}}],
                [
                    [
                        ["mess:1:0:c0:x86_64", "gtk:3:0:c8:x86_64", "platform:f28:0:c0:x86_64"],
                        ["mess:1:0:c0:x86_64", "gtk:4:0:c9:x86_64", "platform:f29:0:c0:x86_64"],
                    ]
                ],
            ),
        ),
    )
    def test_solve(self, dependencies, expected):
        modules = (
            ("platform:f28:0:c0", []),
            ("platform:f29:0:c0", []),
            ("gtk:3:0:c8", [{"requires": {"platform": ["f28"]}}]),
            ("gtk:3:0:c9", [{"requires": {"platform": ["f29"]}}]),
            ("gtk:4:0:c8", [{"requires": {"platform": ["f28"]}}]),
            ("gtk:4:0:c9", [{"requires": {"platform": ["f29"]}}]),
            ("qt:4:0:c8", [{"requires": {"platform": ["f28"]}}]),
            ("qt:4:0:c9", [{"requires": {"platform": ["f29"]}}]),
            ("qt:5:0:c8", [{"requires": {"platform": ["f28"]}}]),
            ("qt:5:0:c9", [{"requires": {"platform": ["f29"]}}]),
            (
                "mess:1:0:c0",
                [
                    {"requires": {"gtk": ["3"], "platform": ["f28"]}},
                    {"requires": {"gtk": ["4"], "platform": ["f29"]}},
                ],
            ),
        )
        for nsvc, deps in modules:
            self.mmd_resolver.add_modules(
                make_module(nsvc, dependencies=deps))

        app = make_module("app:1:0", dependencies=dependencies)
        expanded = self.mmd_resolver.solve(app)

        expected = set(
            frozenset(["app:1:0:%d:src" % c] + e) for c, exp in enumerate(expected) for e in exp
        )

        assert expanded == expected

    @pytest.mark.parametrize(
        "dependencies, expected",
        (
            # BR all platform streams -> build for all platform streams.
            (
                [{"buildrequires": {"platform": []}}],
                [
                    [
                        ["platform:el8.2.0.z:0:c0:x86_64"],
                        ["platform:el8.1.0:0:c0:x86_64"],
                        ["platform:el8.0.0:0:c0:x86_64"],
                        ["platform:el7.6.0:0:c0:x86_64"],
                    ]
                ],
            ),
            # BR "el8" platform stream -> build for all el8 platform streams.
            (
                [{"buildrequires": {"platform": ["el8"]}}],
                [
                    [
                        ["platform:el8.2.0.z:0:c0:x86_64"],
                        ["platform:el8.1.0:0:c0:x86_64"],
                        ["platform:el8.0.0:0:c0:x86_64"],
                    ]
                ],
            ),
            # BR "el8.1.0" platform stream -> build just for el8.1.0.
            (
                [{"buildrequires": {"platform": ["el8.1.0"]}}],
                [[["platform:el8.1.0:0:c0:x86_64"]]]
            ),
            # BR platform:el8.1.0 and gtk:3, which is not built against el8.1.0,
            # but it is built only against el8.0.0 -> cherry-pick gtk:3 from el8.0.0
            # and build once against platform:el8.1.0.
            (
                [{"buildrequires": {"platform": ["el8.1.0"], "gtk": ["3"]}}],
                [[["platform:el8.1.0:0:c0:x86_64", "gtk:3:0:c8:x86_64"]]],
            ),
            # BR platform:el8.2.0 and gtk:3, this time gtk:3 build against el8.2.0 exists
            # -> use both platform and gtk from el8.2.0 and build once.
            (
                [{"buildrequires": {"platform": ["el8.2.0.z"], "gtk": ["3"]}}],
                [[["platform:el8.2.0.z:0:c0:x86_64", "gtk:3:1:c8:x86_64"]]],
            ),
            # BR platform:el8.2.0 and mess:1 which is built against platform:el8.1.0 and
            # requires gtk:3 which is built against platform:el8.2.0 and platform:el8.0.0
            # -> Use platform:el8.2.0 and
            # -> cherry-pick mess:1 from el8.1.0 and
            # -> use gtk:3:1 from el8.2.0.
            (
                [{"buildrequires": {"platform": ["el8.2.0.z"], "mess": ["1"]}}],
                [[["platform:el8.2.0.z:0:c0:x86_64", "mess:1:0:c0:x86_64", "gtk:3:1:c8:x86_64"]]],
            ),
            # BR platform:el8.1.0 and mess:1 which is built against platform:el8.1.0 and
            # requires gtk:3 which is built against platform:el8.2.0 and platform:el8.0.0
            # -> Use platform:el8.1.0 and
            # -> Used mess:1 from el8.1.0 and
            # -> cherry-pick gtk:3:0 from el8.0.0.
            (
                [{"buildrequires": {"platform": ["el8.1.0"], "mess": ["1"]}}],
                [[["platform:el8.1.0:0:c0:x86_64", "mess:1:0:c0:x86_64", "gtk:3:0:c8:x86_64"]]],
            ),
            # BR platform:el8.0.0 and mess:1 which is built against platform:el8.1.0 and
            # requires gtk:3 which is built against platform:el8.2.0 and platform:el8.0.0
            # -> No valid combination, because mess:1 is only available in el8.1.0 and later.
            ([{"buildrequires": {"platform": ["el8.0.0"], "mess": ["1"]}}], []),
            # This is undefined... it might build just once against latest platform or
            # against all the platforms... we don't know
            # ({"platform": ["el8"], "gtk": ["3"]}, {}, [
            #     [["platform:el8.2.0:0:c0:x86_64", "gtk:3:1:c8:x86_64"]],
            # ]),
        ),
    )
    def test_solve_virtual_streams(self, dependencies, expected):
        modules = (
            # (nsvc, dependencies, xmd_mbs_buildrequires, virtual_streams)
            ("platform:el8.0.0:0:c0", [], {}, ["el8"]),
            ("platform:el8.1.0:0:c0", [], {}, ["el8"]),
            ("platform:el8.2.0.z:0:c0", [], {}, ["el8"]),
            ("platform:el7.6.0:0:c0", [], {}, ["el7"]),
            (
                "gtk:3:0:c8",
                [{"requires": {"platform": ["el8"]}}],
                {"platform": {"stream": "el8.0.0"}},
                None
            ),
            (
                "gtk:3:1:c8",
                [{"requires": {"platform": ["el8"]}}],
                {"platform": {"stream": "el8.2.0.z"}},
                None
            ),
            (
                "mess:1:0:c0",
                [{"requires": {"gtk": ["3"], "platform": ["el8"]}}],
                {"platform": {"stream": "el8.1.0"}},
                None
            ),
        )
        for nsvc, deps, xmd_mbs_brs, virtual_streams in modules:
            xmd = {"mbs": {"buildrequires": xmd_mbs_brs}}
            if virtual_streams:
                xmd["mbs"]["virtual_streams"] = virtual_streams
            self.mmd_resolver.add_modules(
                make_module(nsvc, dependencies=deps, xmd=xmd))

        app = make_module("app:1:0", dependencies=dependencies)
        if not expected:
            with pytest.raises(RuntimeError):
                self.mmd_resolver.solve(app)
            return
        else:
            expanded = self.mmd_resolver.solve(app)

        expected = set(
            frozenset(["app:1:0:%d:src" % c] + e) for c, exp in enumerate(expected) for e in exp)

        assert expanded == expected

    @pytest.mark.parametrize(
        "dependencies, modules, err_msg_regex",
        (
            # app --br--> gtk:1 --req--> bar:1* ---req---> platform:f29
            #    \--br--> foo:1 --req--> bar:2* ---req--/
            (
                [{"buildrequires": {"gtk": ["1"], "foo": ["1"]}}],
                (
                    ("platform:f29:0:c0", []),
                    ("gtk:1:1:c01", [{"requires": {"bar": ["1"]}}]),
                    ("bar:1:0:c02", [{"requires": {"platform": ["f29"]}}]),
                    ("foo:1:1:c03", [{"requires": {"bar": ["2"]}}]),
                    ("bar:2:0:c04", [{"requires": {"platform": ["f29"]}}]),
                ),
                "bar:1:0:c02 and bar:2:0:c04",
            ),
            # app --br--> gtk:1 --req--> bar:1* ----------req----------> platform:f29
            #    \--br--> foo:1 --req--> baz:1 --req--> bar:2* --req--/
            (
                [{"buildrequires": {"gtk": ["1"], "foo": ["1"]}}],
                (
                    ("platform:f29:0:c0", []),
                    ("gtk:1:1:c01", [{"requires": {"bar": ["1"]}}]),
                    ("bar:1:0:c02", [{"requires": {"platform": ["f29"]}}]),
                    ("foo:1:1:c03", [{"requires": {"baz": ["1"]}}]),
                    ("baz:1:1:c04", [{"requires": {"bar": ["2"]}}]),
                    ("bar:2:0:c05", [{"requires": {"platform": ["f29"]}}]),
                ),
                "bar:1:0:c02 and bar:2:0:c05",
            ),
            # Test multiple conflicts pairs are detected.
            # app --br--> gtk:1 --req--> bar:1* ---------req-----------\
            #    \--br--> foo:1 --req--> baz:1 --req--> bar:2* ---req---> platform:f29
            #    \--br--> pkga:1 --req--> perl:5' -------req-----------/
            #    \--br--> pkgb:1 --req--> perl:6' -------req-----------/
            (
                [{
                    "buildrequires": {
                        "gtk": ["1"],
                        "foo": ["1"],
                        "pkga": ["1"],
                        "pkgb": ["1"],
                    }
                }],
                (
                    ("platform:f29:0:c0", {}),
                    ("gtk:1:1:c01", [{"requires": {"bar": ["1"]}}]),
                    ("bar:1:0:c02", [{"requires": {"platform": ["f29"]}}]),
                    ("foo:1:1:c03", [{"requires": {"baz": ["1"]}}]),
                    ("baz:1:1:c04", [{"requires": {"bar": ["2"]}}]),
                    ("bar:2:0:c05", [{"requires": {"platform": ["f29"]}}]),
                    ("pkga:1:0:c06", [{"requires": {"perl": ["5"]}}]),
                    ("perl:5:0:c07", [{"requires": {"platform": ["f29"]}}]),
                    ("pkgb:1:0:c08", [{"requires": {"perl": ["6"]}}]),
                    ("perl:6:0:c09", [{"requires": {"platform": ["f29"]}}]),
                ),
                # MMD Resolver should still catch a conflict
                "bar:1:0:c02 and bar:2:0:c05",
            ),
        ),
    )
    def test_solve_stream_conflicts(self, dependencies, modules, err_msg_regex):
        for nsvc, deps in modules:
            self.mmd_resolver.add_modules(
                make_module(nsvc, dependencies=deps))

        app = make_module("app:1:0", dependencies=dependencies)

        with pytest.raises(RuntimeError, match=err_msg_regex):
            self.mmd_resolver.solve(app)

    def test_solve_new_platform(self,):
        """
        Tests that MMDResolver works in case we add new platform and there is
        modular dependency of input module which is built only against old
        platforms and in the same time the input module wants to be built
        with all available platforms.
        """
        modules = (
            ("platform:f28:0:c0", []),
            ("platform:f29:0:c0", []),
            ("platform:f30:0:c0", []),
            ("gtk:3:0:c8", [{"requires": {"platform": ["f28"]}}]),
            ("gtk:3:0:c9", [{"requires": {"platform": ["f29"]}}]),
        )
        for nsvc, deps in modules:
            self.mmd_resolver.add_modules(
                make_module(nsvc, dependencies=deps))

        app = make_module("app:1:0", dependencies=[
            {"buildrequires": {"platform": [], "gtk": ["3"]}}
        ])
        expanded = self.mmd_resolver.solve(app)

        # Build only against f28 and f29, because "gtk:3" is not built against f30.
        expected = {
            frozenset(["gtk:3:0:c8:x86_64", "app:1:0:0:src", "platform:f28:0:c0:x86_64"]),
            frozenset(["gtk:3:0:c9:x86_64", "app:1:0:0:src", "platform:f29:0:c0:x86_64"]),
        }

        assert expanded == expected

    def test_solve_requires_any_platform(self,):
        """
        Tests that MMDResolver finds out the buildrequired module `foo` even if
        it was built on newer platform stream, but can run on any platform stream.
        """
        modules = (
            ("platform:f28:0:c0", [], {}),
            ("platform:f29:0:c0", [], {}),
            ("platform:f30:0:c0", [], {}),
            (
                "foo:1:0:c8",
                [{"requires": {"platform": []}}],
                {"platform": {"stream": "f29"}}
            ),
        )
        for nsvc, deps, xmd_mbs_brs in modules:
            mmd = make_module(nsvc, dependencies=deps, xmd={
                "mbs": {"buildrequires": xmd_mbs_brs}
            })
            self.mmd_resolver.add_modules(mmd)

        app = make_module("app:1:0", [
            {"buildrequires": {"platform": ["f28"], "foo": ["1"]}}
        ])
        expanded = self.mmd_resolver.solve(app)

        expected = {
            frozenset(["foo:1:0:c8:x86_64", "app:1:0:0:src", "platform:f28:0:c0:x86_64"]),
        }

        assert expanded == expected

    @pytest.mark.parametrize(
        "nsvc, dependencies, expected",
        (
            ("platform:f28:0:c0", [], True),
            ("platform:latest:5:c8", [], False),
            ("gtk:3:0:c8", [{"requires": {"platform": ["f28"]}}], False)
        ),
    )
    def test_base_module_stream_version(self, nsvc, dependencies, expected):
        """
        Tests that add_base_module_provides returns True for base modules with stream versions
        """
        mmd = make_module(nsvc, dependencies=dependencies)
        solvable = self.mmd_resolver.available_repo.add_solvable()
        solvable.name = nsvc
        solvable.evr = str(mmd.get_version())
        solvable.arch = "x86_64"
        assert self.mmd_resolver._add_base_module_provides(solvable, mmd) is expected

    @pytest.mark.parametrize(
        "nsvc, expected",
        (
            ("platform:f28:3:c0", {"module(platform)", "module(platform:f28) = 28.0"}),
            ("platform:latest:5:c8", {"module(platform)", "module(platform:latest) = 5"}),
        ),
    )
    def test_base_module_provides(self, nsvc, expected):
        self.mmd_resolver.add_modules(make_module(nsvc))
        ns = nsvc.rsplit(":", 2)[0]
        provides = self.mmd_resolver.solvables[ns][0].lookup_deparray(solv.SOLVABLE_PROVIDES)
        assert {str(provide) for provide in provides} == expected



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT
from __future__ import absolute_import
import os

from datetime import datetime
from mock import patch, PropertyMock
import pytest

from module_build_service.builder.MockModuleBuilder import load_local_builds
from module_build_service.common import models
from module_build_service.common.config import conf
from module_build_service.common.errors import UnprocessableEntity
from module_build_service.common.models import ModuleBuild
from module_build_service.common.modulemd import Modulemd
from module_build_service.common.utils import import_mmd, load_mmd, mmd_to_str
import module_build_service.resolver as mbs_resolver
from module_build_service.scheduler.db_session import db_session
import tests


class TestDBModule:

    def test_get_buildrequired_modulemds(self):
        mmd = load_mmd(tests.read_staged_data("platform"))
        mmd = mmd.copy(mmd.get_module_name(), "f30.1.3")

        import_mmd(db_session, mmd)
        platform_f300103 = db_session.query(ModuleBuild).filter_by(stream="f30.1.3").one()
        mmd = tests.make_module("testmodule:master:20170109091357:123")
        build = ModuleBuild(
            name="testmodule",
            stream="master",
            version=20170109091357,
            state=5,
            build_context="dd4de1c346dcf09ce77d38cd4e75094ec1c08ec3",
            runtime_context="ec4de1c346dcf09ce77d38cd4e75094ec1c08ef7",
            context="7c29193d",
            koji_tag="module-testmodule-master-20170109091357-7c29193d",
            scmurl="https://src.stg.fedoraproject.org/modules/testmodule.git?#ff1ea79",
            batch=3,
            owner="Dr. Pepper",
            time_submitted=datetime(2018, 11, 15, 16, 8, 18),
            time_modified=datetime(2018, 11, 15, 16, 19, 35),
            rebuild_strategy="changed-and-after",
            modulemd=mmd_to_str(mmd),
        )
        build.buildrequires.append(platform_f300103)
        db_session.add(build)
        db_session.commit()

        resolver = mbs_resolver.GenericResolver.create(db_session, conf, backend="db")
        result = resolver.get_buildrequired_modulemds(
            "testmodule", "master", platform_f300103.mmd())
        nsvcs = {m.get_nsvc() for m in result}
        assert nsvcs == {"testmodule:master:20170109091357:123"}

    @pytest.mark.parametrize("stream_versions", [False, True])
    @pytest.mark.parametrize("provide_test_data",
                             [{"multiple_stream_versions": True}], indirect=True)
    def test_get_compatible_base_module_modulemds_stream_versions(self, stream_versions,
                                                                  provide_test_data):
        resolver = mbs_resolver.GenericResolver.create(db_session, conf, backend="db")
        platform = db_session.query(ModuleBuild).filter_by(name="platform", stream="f29.1.0").one()
        platform_mmd = platform.mmd()
        result = resolver.get_compatible_base_module_modulemds(
            platform_mmd, stream_version_lte=stream_versions, virtual_streams=["f29"],
            states=[models.BUILD_STATES["ready"]])
        nsvcs = {mmd.get_nsvc() for mmd in result}
        if stream_versions:
            assert nsvcs == {"platform:f29.1.0:3:00000000", "platform:f29.0.0:3:00000000"}
        else:
            assert nsvcs == {
                "platform:f29.1.0:3:00000000",
                "platform:f29.0.0:3:00000000",
                "platform:f29.2.0:3:00000000"
            }

    @pytest.mark.parametrize("empty_buildrequires", [False, True])
    def test_get_module_build_dependencies(self, empty_buildrequires, reuse_component_init_data):
        """
        Tests that the buildrequires of testmodule are returned
        """
        expected = {"module-f28-build"}
        module = models.ModuleBuild.get_by_id(db_session, 2)
        if empty_buildrequires:
            expected = set()
            module = models.ModuleBuild.get_by_id(db_session, 2)
            mmd = module.mmd()
            # Wipe out the dependencies
            mmd.clear_dependencies()
            xmd = mmd.get_xmd()
            xmd["mbs"]["buildrequires"] = {}
            mmd.set_xmd(xmd)
            module.modulemd = mmd_to_str(mmd)
            db_session.commit()
        resolver = mbs_resolver.GenericResolver.create(db_session, conf, backend="db")
        result = resolver.get_module_build_dependencies(
            "testmodule", "master", "20170109091357", "78e4a6fd").keys()
        assert set(result) == expected

    def test_get_module_build_dependencies_recursive(self, reuse_component_init_data):
        """
        Tests that the buildrequires are returned when it is two layers deep
        """
        # Add testmodule2 that requires testmodule
        module = models.ModuleBuild.get_by_id(db_session, 3)
        mmd = module.mmd()
        # Rename the module
        mmd = mmd.copy("testmodule2")
        mmd.set_version(20180123171545)
        deps = Modulemd.Dependencies()
        deps.add_runtime_stream("testmodule", "master")
        mmd.add_dependencies(deps)
        xmd = mmd.get_xmd()
        xmd["mbs"]["requires"]["testmodule"] = {
            "filtered_rpms": [],
            "ref": "620ec77321b2ea7b0d67d82992dda3e1d67055b4",
            "stream": "master",
            "version": "20180205135154",
        }
        mmd.set_xmd(xmd)
        module.modulemd = mmd_to_str(mmd)
        module.name = "testmodule2"
        module.version = str(mmd.get_version())
        module.koji_tag = "module-ae2adf69caf0e1b6"

        db_session.commit()

        resolver = mbs_resolver.GenericResolver.create(db_session, conf, backend="db")
        result = resolver.get_module_build_dependencies(
            "testmodule2", "master", "20180123171545", "c40c156c").keys()
        assert set(result) == {"module-f28-build"}

    @patch(
        "module_build_service.common.config.Config.system",
        new_callable=PropertyMock,
        return_value="test",
    )
    @patch(
        "module_build_service.common.config.Config.mock_resultsdir",
        new_callable=PropertyMock,
        return_value=tests.staged_data_filename("local_builds"),
    )
    def test_get_module_build_dependencies_recursive_requires(self, resultdir, conf_system,
                                                              reuse_component_init_data):
        """
        Tests that it returns the requires of the buildrequires recursively
        """
        load_local_builds(["platform:f30", "parent", "child", "testmodule"])

        build = models.ModuleBuild.local_modules(db_session, "child", "master")
        resolver = mbs_resolver.GenericResolver.create(db_session, conf, backend="db")
        result = resolver.get_module_build_dependencies(mmd=build[0].mmd()).keys()

        local_path = tests.staged_data_filename("local_builds")

        expected = [os.path.join(local_path, "module-parent-master-20170816080815/results")]
        assert set(result) == set(expected)

    def test_resolve_requires(self):
        build = models.ModuleBuild.get_by_id(db_session, 2)
        resolver = mbs_resolver.GenericResolver.create(db_session, conf, backend="db")
        result = resolver.resolve_requires(
            [":".join([build.name, build.stream, build.version, build.context])]
        )

        assert result == {
            "testmodule": {
                "stream": "master",
                "version": "20170109091357",
                "context": u"78e4a6fd",
                "ref": "ff1ea79fc952143efeed1851aa0aa006559239ba",
                "koji_tag": "module-testmodule-master-20170109091357-78e4a6fd",
            }
        }

    def test_resolve_requires_exception(self, reuse_component_init_data):
        build = models.ModuleBuild.get_by_id(db_session, 2)
        resolver = mbs_resolver.GenericResolver.create(db_session, conf, backend="db")
        with pytest.raises(UnprocessableEntity):
            resolver.resolve_requires(
                [":".join(["abcdefghi", build.stream, build.version, build.context])]
            )

    def test_resolve_requires_siblings(self, require_platform_and_default_arch):
        resolver = mbs_resolver.GenericResolver.create(db_session, conf, backend="db")
        mmd = load_mmd(tests.read_staged_data("formatted_testmodule"))
        for i in range(3):
            build = tests.module_build_from_modulemd(mmd_to_str(mmd))
            build.context = "f6e2ae" + str(i)
            build.build_context = "f6e2aeec7576196241b9afa0b6b22acf2b6873d" + str(i)
            build.runtime_context = "bbc84c7b817ab3dd54916c0bcd6c6bdf512f7f9c" + str(i)
            build.state = models.BUILD_STATES["ready"]
            db_session.add(build)
        db_session.commit()

        build_one = ModuleBuild.get_by_id(db_session, 2)
        nsvc = ":".join([build_one.name, build_one.stream, build_one.version, build_one.context])
        result = resolver.resolve_requires([nsvc])
        assert result == {
            "testmodule": {
                "stream": build_one.stream,
                "version": build_one.version,
                "context": build_one.context,
                "ref": "65a7721ee4eff44d2a63fb8f3a8da6e944ab7f4d",
                "koji_tag": None
            }
        }

        db_session.commit()

    def test_resolve_profiles(self):
        """
        Tests that the profiles get resolved recursively
        """
        mmd = models.ModuleBuild.get_by_id(db_session, 2).mmd()
        resolver = mbs_resolver.GenericResolver.create(db_session, conf, backend="db")
        result = resolver.resolve_profiles(mmd, ("buildroot", "srpm-buildroot"))
        expected = {
            "buildroot": {
                "unzip",
                "tar",
                "cpio",
                "gawk",
                "gcc",
                "xz",
                "sed",
                "findutils",
                "util-linux",
                "bash",
                "info",
                "bzip2",
                "grep",
                "redhat-rpm-config",
                "fedora-release",
                "diffutils",
                "make",
                "patch",
                "shadow-utils",
                "coreutils",
                "which",
                "rpm-build",
                "gzip",
                "gcc-c++",
            },
            "srpm-buildroot": {
                "shadow-utils",
                "redhat-rpm-config",
                "rpm-build",
                "fedora-release",
                "fedpkg-minimal",
                "gnupg2",
                "bash",
            },
        }
        assert result == expected

    @patch(
        "module_build_service.common.config.Config.system",
        new_callable=PropertyMock,
        return_value="test",
    )
    @patch(
        "module_build_service.common.config.Config.mock_resultsdir",
        new_callable=PropertyMock,
        return_value=tests.staged_data_filename("local_builds")
    )
    def test_resolve_profiles_local_module(self, local_builds, conf_system):
        """
        Test that profiles get resolved recursively on local builds
        """
        # This test requires a platform module loaded from local rather than
        # the one added to database.
        platform = db_session.query(models.ModuleBuild).filter(
            models.ModuleBuild.name == "platform"
        ).one()
        db_session.delete(platform)
        db_session.commit()

        load_local_builds(["platform:f28"])
        mmd = models.ModuleBuild.get_by_id(db_session, 2).mmd()
        resolver = mbs_resolver.GenericResolver.create(db_session, conf, backend="mbs")
        result = resolver.resolve_profiles(mmd, ("buildroot", "srpm-buildroot"))
        expected = {"buildroot": {"foo"}, "srpm-buildroot": {"bar"}}
        assert result == expected

    @pytest.mark.parametrize("provide_test_data",
                             [{"multiple_stream_versions": True}], indirect=True)
    def test_get_latest_with_virtual_stream(self, provide_test_data):
        resolver = mbs_resolver.GenericResolver.create(db_session, conf, backend="db")
        mmd = resolver.get_latest_with_virtual_stream("platform", "f29")
        assert mmd
        assert mmd.get_stream_name() == "f29.2.0"

    def test_get_latest_with_virtual_stream_none(self):
        resolver = mbs_resolver.GenericResolver.create(db_session, conf, backend="db")
        mmd = resolver.get_latest_with_virtual_stream("platform", "doesnotexist")
        assert not mmd

    def test_get_module_count(self):
        resolver = mbs_resolver.GenericResolver.create(db_session, conf, backend="db")
        count = resolver.get_module_count(name="platform", stream="f28")
        assert count == 1



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT

"""
This module defines constant for events emitted by external services that work
with MBS together to complete a module build.

The event name is defined in general as much as possible, especially for the
events from Koji. Because some instance based on Koji, like Brew, might send
messages to different topics on different message bus. For example, when a
build is complete, Koji sends a message to topic buildsys.build.state.change,
however Brew sends to topic brew.build.complete, etc.
"""

from __future__ import absolute_import
from functools import wraps
import sched
import time

from module_build_service.common import log


KOJI_BUILD_CHANGE = "koji_build_change"
KOJI_TAG_CHANGE = "koji_tag_change"
KOJI_REPO_CHANGE = "koji_repo_change"
MBS_MODULE_STATE_CHANGE = "mbs_module_state_change"
GREENWAVE_DECISION_UPDATE = "greenwave_decision_update"


class Scheduler(sched.scheduler):
    """
    Subclass of `sched.scheduler` allowing to schedule handlers calls.

    If one of the MBS handler functions need to call another handler, they need to do it in a safe
    way - such another handler call should not be done in the middle of another handler's
    execution.

    This class provides an solution for that. Handler can schedule run of other handler using
    the `add` method. The handlers should use `mbs_event_handler` decorator which ensures that
    the `run` method is called at the end of handler's execution and other scheduler handlers
    are executed.
    """

    def add(self, handler, arguments=()):
        """
        Schedule execution of `handler` with `arguments`.
        """
        self.enter(0, 0, handler.delay, arguments)

    def run(self):
        """
        Runs scheduled handlers.
        """
        log.debug("Running event scheduler with following events:")
        for event in self.queue:
            log.debug("    %r", event)
        sched.scheduler.run(self)

    def reset(self):
        """
        Resets the Scheduler to initial state.
        """
        while not self.empty():
            self.cancel(self.queue[0])


scheduler = Scheduler(time.time, delayfunc=lambda x: x)


def mbs_event_handler(func):
    """
    A decorator for MBS event handlers. It implements common tasks which should otherwise
    be repeated in every MBS event handler, for example:

      - at the end of handler, call events.scheduler.run().
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        finally:
            scheduler.run()
    # save origin function as functools.wraps from python2 doesn't preserve the signature
    if not hasattr(wrapper, "__wrapped__"):
        wrapper.__wrapped__ = func
    return wrapper



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT
from __future__ import absolute_import

from mock import patch, MagicMock
import pytest

from module_build_service.common.errors import IgnoreMessage
from module_build_service.scheduler import events
from module_build_service.scheduler.consumer import MBSConsumer


class TestConsumer:
    def test_get_abstracted_msg_fedmsg(self):
        """
        Test the output of get_abstracted_msg() when using the
        fedmsg backend.
        """
        hub = MagicMock(config={})
        consumer = MBSConsumer(hub)
        msg = {
            "username": "apache",
            "source_name": "datanommer",
            "i": 1,
            "timestamp": 1505492681.0,
            "msg_id": "2017-0627b798-f241-4230-b365-8a8a111a8ec5",
            "crypto": "x509",
            "topic": "org.fedoraproject.prod.buildsys.tag",
            "headers": {},
            "source_version": "0.8.1",
            "msg": {
                "build_id": 962861,
                "name": "python3-virtualenv",
                "tag_id": 263,
                "instance": "primary",
                "tag": "epel7-pending",
                "user": "bodhi",
                "version": "15.1.0",
                "owner": "orion",
                "release": "1.el7",
            },
        }
        event_info = consumer.get_abstracted_event_info(msg)
        assert event_info["event"] == events.KOJI_TAG_CHANGE
        assert event_info["msg_id"] == msg["msg_id"]
        assert event_info["tag_name"] == msg["msg"]["tag"]

    @patch("module_build_service.scheduler.consumer.models")
    @patch.object(MBSConsumer, "process_message")
    def test_consume_fedmsg(self, process_message, models):
        """
        Test the MBSConsumer.consume() method when using the
        fedmsg backend.
        """
        hub = MagicMock(config={})
        consumer = MBSConsumer(hub)
        msg = {
            "topic": "org.fedoraproject.prod.buildsys.repo.done",
            "headers": {},
            "body": {
                "username": "apache",
                "source_name": "datanommer",
                "i": 1,
                "timestamp": 1405126329.0,
                "msg_id": "2014-adbc33f6-51b0-4fce-aa0d-3c699a9920e4",
                "crypto": "x509",
                "topic": "org.fedoraproject.prod.buildsys.repo.done",
                "headers": {},
                "source_version": "0.6.4",
                "msg": {
                    "instance": "primary",
                    "repo_id": 400859,
                    "tag": "f22-build",
                    "tag_id": 278,
                },
            },
        }
        consumer.consume(msg)
        assert process_message.call_count == 1
        event_info = process_message.call_args[0][0]
        assert event_info["event"] == events.KOJI_REPO_CHANGE
        assert event_info["msg_id"] == msg["body"]["msg_id"]
        assert event_info["tag_name"] == msg["body"]["msg"]["tag"]

    @patch.object(MBSConsumer, "process_message")
    def test_ingore_koji_build_change_event_without_task_id(self, process_message):
        """
        Test koji_build_change event without task_id should be ignored.
        """
        hub = MagicMock(config={})
        consumer = MBSConsumer(hub)
        event = {
            'build_new_state': 1,
            'task_id': None,
            'msg_id': u'f66a43be-e510-44fc-a318-e422cfda65d3',
            'module_build_id': None,
            'state_reason': None,
            'build_name': 'foobar',
            'build_version': '201912130626',
            'event': 'koji_build_change',
            'build_release': u'070752'
        }
        consumer.get_abstracted_event_info = MagicMock()
        consumer.get_abstracted_event_info.return_value = event
        consumer.consume({})
        assert process_message.call_count == 0

    def test_validate_event_none_msg(self):
        hub = MagicMock(config={})
        consumer = MBSConsumer(hub)
        with pytest.raises(IgnoreMessage):
            consumer.validate_event(None)



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT
""" Handlers for repo change events on the message bus. """

from __future__ import absolute_import
from datetime import datetime
import logging

from module_build_service.builder import GenericBuilder
from module_build_service.common import conf, log, models
from module_build_service.scheduler import celery_app, events
from module_build_service.scheduler.batches import start_next_batch_build
from module_build_service.scheduler.db_session import db_session

logging.basicConfig(level=logging.DEBUG)


@celery_app.task
@events.mbs_event_handler
def done(msg_id, tag_name):
    """Called whenever koji rebuilds a repo, any repo.

    :param str msg_id: the original id of the message being handled which is
        received from the message bus.
    :param str tag_name: the tag name from which the repo is generated.
    """

    # First, find our ModuleBuild associated with this repo, if any.
    if conf.system in ("koji", "test") and not tag_name.endswith("-build"):
        log.debug("Tag %r does not end with '-build' suffix, ignoring", tag_name)
        return
    tag = tag_name[:-6] if tag_name.endswith("-build") else tag_name
    module_build = models.ModuleBuild.get_by_tag(db_session, tag_name)
    if not module_build:
        log.debug("No module build found associated with koji tag %r" % tag)
        return

    # It is possible that we have already failed.. but our repo is just being
    # routinely regenerated.  Just ignore that.  If module_build_service says the module is
    # dead, then the module is dead.
    if module_build.state == models.BUILD_STATES["failed"]:
        log.info("Ignoring repo regen for already failed %r" % module_build)
        return

    # If there are no components in this module build, then current_batch will be empty
    if module_build.component_builds:
        current_batch = module_build.current_batch()
    else:
        current_batch = []

    # Get the list of untagged components in current/previous batches which
    # have been built successfully
    if conf.system in ("koji", "test") and current_batch:
        if any(c.is_completed and not c.is_tagged for c in module_build.up_to_current_batch()):
            log.info("Ignoring repo regen, because not all components are tagged.")
            return
        if all(c.is_waiting_for_build for c in current_batch):
            log.info("Ignoring repo regen because no components have started in the batch.")
            return

    # If any in the current batch are still running.. just wait.
    running = [c for c in current_batch if c.is_building]
    if running:
        log.info(
            "%r has %r of %r components still building in this batch (%r total)",
            module_build, len(running), len(current_batch), len(module_build.component_builds)
        )
        return

    # Assemble the list of all successful components in the batch.
    good = [c for c in current_batch if c.is_completed]

    # If *none* of the components completed for this batch, then obviously the
    # module fails.  However!  We shouldn't reach this scenario.  There is
    # logic over in the component handler which should fail the module build
    # first before we ever get here.  This is here as a race condition safety
    # valve.
    if module_build.component_builds and not good:
        state_reason = "Component(s) {} failed to build.".format(
            ", ".join(c.package for c in current_batch if c.is_unsuccessful))
        module_build.transition(
            db_session, conf, models.BUILD_STATES["failed"], state_reason, failure_type="infra")
        db_session.commit()
        log.warning("Odd!  All components in batch failed for %r." % module_build)
        return

    groups = GenericBuilder.default_buildroot_groups(db_session, module_build)
    builder = GenericBuilder.create(
        db_session,
        module_build.owner,
        module_build,
        conf.system,
        conf,
        tag_name=tag,
        components=[c.package for c in module_build.component_builds],
    )
    builder.buildroot_connect(groups)

    # If we have reached here then we know the following things:
    #
    # - All components in this batch have finished (failed or succeeded)
    # - One or more succeeded.
    # - They have been regenerated back into the buildroot.
    #
    # So now we can either start a new batch if there are still some to build
    # or, if everything is built successfully, then we can bless the module as
    # complete.
    has_unbuilt_components = any(c.is_unbuilt for c in module_build.component_builds)
    has_failed_components = any(c.is_unsuccessful for c in module_build.component_builds)

    if has_unbuilt_components and not has_failed_components:
        # Ok, for the subset of builds that did complete successfully, check to
        # see if they are in the buildroot before starting new batch.
        artifacts = [component_build.nvr for component_build in good]
        if not builder.buildroot_ready(artifacts):
            log.info("Not all of %r are in the buildroot.  Waiting." % artifacts)
            return

        # Try to start next batch build, because there are still unbuilt
        # components in a module.
        start_next_batch_build(conf, module_build, builder)
    else:
        if has_failed_components:
            state_reason = "Component(s) {} failed to build.".format(
                ", ".join(
                    c.package for c in module_build.component_builds if c.is_unsuccessful
                )
            )
            module_build.transition(
                db_session,
                conf,
                state=models.BUILD_STATES["failed"],
                state_reason=state_reason,
                failure_type="user",
            )
        else:
            # Tell the external buildsystem to wrap up (CG import, createrepo, etc.)
            module_build.time_completed = datetime.utcnow()
            builder.finalize(succeeded=True)

            module_build.transition(db_session, conf, state=models.BUILD_STATES["done"])
        db_session.commit()



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT
"""
Allows MBS to run behind reverse proxy and to ensure redirects work with https.

WSGI Middleware!!

Source: http://flask.pocoo.org/snippets/35/ by Peter Hansen
"""
from __future__ import absolute_import


class ReverseProxy(object):
    """Wrap the application in this middleware and configure the
    front-end server to add these headers, to let you quietly bind
    this to a URL other than / and to an HTTP scheme that is
    different than what is used locally.

    :param app: the WSGI application
    """

    def __init__(self, app):
        self.app = app

    def __call__(self, environ, start_response):
        script_name = environ.get("HTTP_X_SCRIPT_NAME", "")
        if script_name:
            environ["SCRIPT_NAME"] = script_name
            path_info = environ["PATH_INFO"]
            if path_info.startswith(script_name):
                environ["PATH_INFO"] = path_info[len(script_name):]

        server = environ.get("HTTP_X_FORWARDED_HOST", "")
        if server:
            environ["HTTP_HOST"] = server

        scheme = environ.get("HTTP_X_SCHEME", "")
        if scheme:
            environ["wsgi.url_scheme"] = scheme
        return self.app(environ, start_response)



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT
from __future__ import absolute_import

from module_build_service.common.config import conf
from module_build_service.common.errors import StreamAmbigous
from module_build_service.resolver import GenericResolver


def expand_single_mse_streams(
        db_session, name, streams, default_streams=None, raise_if_stream_ambigous=False):
    """
    Helper method for `expand_mse_stream()` expanding single name:[streams].
    Returns list of expanded streams.

    :param db_session: SQLAlchemy DB session.
    :param str name: Name of the module which will be expanded.
    :param streams: List of streams to expand.
    :type streams: list[str]
    :param dict default_streams: Dict in {module_name: module_stream, ...} format defining
        the default stream to choose for module in case when there are multiple streams to
        choose from.
    :param bool raise_if_stream_ambigous: When True, raises a StreamAmbigous exception in case
        there are multiple streams for some dependency of module and the module name is not
        defined in `default_streams`, so it is not clear which stream should be used.
    """
    from module_build_service.common import models

    default_streams = default_streams or {}
    # Stream can be prefixed with '-' sign to define that this stream should
    # not appear in a resulting list of streams. There can be two situations:
    # a) all streams have '-' prefix. In this case, we treat list of streams
    #    as blacklist and we find all the valid streams and just remove those with
    #    '-' prefix.
    # b) there is at least one stream without '-' prefix. In this case, we can
    #    ignore all the streams with '-' prefix and just add those without
    #    '-' prefix to the list of valid streams.
    streams_is_blacklist = all(stream.startswith("-") for stream in streams)
    if streams_is_blacklist or len(streams) == 0:
        if name in default_streams:
            expanded_streams = [default_streams[name]]
        elif raise_if_stream_ambigous:
            raise StreamAmbigous("There are multiple streams to choose from for module %s." % name)
        else:
            builds = models.ModuleBuild.get_last_build_in_all_streams(db_session, name)
            expanded_streams = [build.stream for build in builds]
    else:
        expanded_streams = []
    for stream in streams:
        if stream.startswith("-"):
            if streams_is_blacklist and stream[1:] in expanded_streams:
                expanded_streams.remove(stream[1:])
        else:
            expanded_streams.append(stream)

    if len(expanded_streams) > 1:
        if name in default_streams:
            expanded_streams = [default_streams[name]]
        elif raise_if_stream_ambigous:
            raise StreamAmbigous(
                "There are multiple streams %r to choose from for module %s."
                % (expanded_streams, name)
            )

    return expanded_streams


def get_compatible_base_module_mmds(resolver, base_mmd, ignore_ns=None):
    """
    Returns dict containing the base modules compatible with `base_mmd` grouped by their state.

    :param GenericResolver resolver: GenericResolver instance.
    :param Modulemd base_mmd: Modulemd instant to return compatible modules for.
    :param set ignore_ns: When set, defines the name:stream of base modules which will be ignored
        by this function and therefore not returned.
    :return dict: Dictionary with module's state name as a key and list of Modulemd objects for
        each compatible base module in that state. For example:
            {
                "ready": [base_mmd_1, base_mmd_2]
                "garbage": [base_mmd_3]
            }
        The input `base_mmd` is always included in the result in "ready" state.
    """
    from module_build_service.common import models

    # Add the module to `seen` and `ret`.
    ret = {"ready": [], "garbage": []}
    ret["ready"].append(base_mmd)
    ns = ":".join([base_mmd.get_module_name(), base_mmd.get_stream_name()])
    seen = set() if not ignore_ns else set(ignore_ns)
    seen.add(ns)

    # Get the list of compatible virtual streams.
    xmd = base_mmd.get_xmd()
    virtual_streams = xmd.get("mbs", {}).get("virtual_streams")

    # In case there are no virtual_streams in the buildrequired name:stream,
    # it is clear that there are no compatible streams, so return just this
    # `base_mmd`.
    if not virtual_streams:
        return ret

    if conf.allow_only_compatible_base_modules:
        stream_version_lte = True
        states = ["ready"]
    else:
        stream_version_lte = False
        states = ["ready", "garbage"]

    for state in states:
        mmds = resolver.get_compatible_base_module_modulemds(
            base_mmd, stream_version_lte, virtual_streams,
            [models.BUILD_STATES[state]])
        ret_chunk = []
        # Add the returned mmds to the `seen` set to avoid querying those
        # individually if they are part of the buildrequire streams for this
        # base module.
        for mmd_ in mmds:
            mmd_ns = ":".join([mmd_.get_module_name(), mmd_.get_stream_name()])
            # An extra precaution to ensure there are no duplicates. This can happen when there
            # are two modules with the same name:stream - one in ready state and one in garbage
            # state.
            if mmd_ns not in seen:
                ret_chunk.append(mmd_)
                seen.add(mmd_ns)
        ret[state] += ret_chunk

    return ret


def get_base_module_mmds(db_session, mmd):
    """
    Returns list of MMDs of base modules buildrequired by `mmd` including the compatible
    old versions of the base module based on the stream version.

    :param Modulemd mmd: Input modulemd metadata.
    :rtype: dict with lists of Modulemd
    :return: Dict with "ready" or "garbage" state name as a key and list of MMDs of base modules
        buildrequired by `mmd` as a value.
    """
    from module_build_service.common import models

    seen = set()
    ret = {"ready": [], "garbage": []}

    resolver = GenericResolver.create(db_session, conf)
    for deps in mmd.get_dependencies():
        buildrequires = {
            module: deps.get_buildtime_streams(module)
            for module in deps.get_buildtime_modules()
        }
        for name in conf.base_module_names:
            if name not in buildrequires.keys():
                continue

            # Since the query below uses stream_version_lte, we can sort the streams by stream
            # version in descending order to not perform unnecessary queries. Otherwise, if the
            # streams are f29.1.0 and f29.2.0, then two queries will occur, causing f29.1.0 to be
            # returned twice. Only one query for that scenario is necessary.
            sorted_desc_streams = sorted(
                buildrequires[name], reverse=True, key=models.ModuleBuild.get_stream_version)
            for stream in sorted_desc_streams:
                ns = ":".join([name, stream])
                if ns in seen:
                    continue

                # Get the MMD for particular buildrequired name:stream to find out the virtual
                # streams according to which we can find the compatible streams later.
                # The returned MMDs are all the module builds for name:stream in the highest
                # version. Given the base module does not depend on other modules, it can appear
                # only in single context and therefore the `mmds` should always contain just
                # zero or one module build.
                mmds = resolver.get_module_modulemds(name, stream)
                if not mmds:
                    continue
                base_mmd = mmds[0]

                new_ret = get_compatible_base_module_mmds(resolver, base_mmd, ignore_ns=seen)
                for state in new_ret.keys():
                    for mmd_ in new_ret[state]:
                        mmd_ns = ":".join([mmd_.get_module_name(), mmd_.get_stream_name()])
                        seen.add(mmd_ns)
                    ret[state] += new_ret[state]
            break
    return ret



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT
from __future__ import absolute_import

import pytest

from module_build_service.common import models
from module_build_service.common.errors import UnprocessableEntity
from module_build_service.common.utils import import_mmd, load_mmd
from module_build_service.scheduler.db_session import db_session
from tests import read_staged_data


@pytest.mark.parametrize("context", ["c1", None])
def test_import_mmd_contexts(context):
    mmd = load_mmd(read_staged_data("formatted_testmodule"))
    mmd.set_context(context)

    xmd = mmd.get_xmd()
    xmd["mbs"]["koji_tag"] = "foo"
    mmd.set_xmd(xmd)

    build, msgs = import_mmd(db_session, mmd)

    mmd_context = build.mmd().get_context()
    if context:
        assert mmd_context == context
        assert build.context == context
    else:
        assert mmd_context == models.DEFAULT_MODULE_CONTEXT
        assert build.context == models.DEFAULT_MODULE_CONTEXT


def test_import_mmd_multiple_dependencies():
    mmd = load_mmd(read_staged_data("formatted_testmodule"))
    mmd.add_dependencies(mmd.get_dependencies()[0].copy())

    expected_error = "The imported module's dependencies list should contain just one element"
    with pytest.raises(UnprocessableEntity) as e:
        import_mmd(db_session, mmd)
        assert str(e.value) == expected_error


def test_import_mmd_no_xmd_buildrequires():
    mmd = load_mmd(read_staged_data("formatted_testmodule"))
    xmd = mmd.get_xmd()
    del xmd["mbs"]["buildrequires"]
    mmd.set_xmd(xmd)

    expected_error = (
        "The imported module buildrequires other modules, but the metadata in the "
        'xmd["mbs"]["buildrequires"] dictionary is missing entries'
    )
    with pytest.raises(UnprocessableEntity) as e:
        import_mmd(db_session, mmd)
        assert str(e.value) == expected_error


def test_import_mmd_minimal_xmd_from_local_repository():
    mmd = load_mmd(read_staged_data("formatted_testmodule"))
    xmd = mmd.get_xmd()
    xmd["mbs"] = {}
    xmd["mbs"]["koji_tag"] = "repofile:///etc/yum.repos.d/fedora-modular.repo"
    xmd["mbs"]["mse"] = True
    xmd["mbs"]["commit"] = "unknown"
    mmd.set_xmd(xmd)

    build, msgs = import_mmd(db_session, mmd, False)
    assert build.name == mmd.get_module_name()


@pytest.mark.parametrize(
    "stream, disttag_marking, error_msg",
    (
        ("f28", None, None),
        ("f28", "fedora28", None),
        ("f-28", "f28", None),
        ("f-28", None, "The stream cannot contain a dash unless disttag_marking is set"),
        ("f28", "f-28", "The disttag_marking cannot contain a dash"),
        ("f-28", "fedora-28", "The disttag_marking cannot contain a dash"),
    ),
)
def test_import_mmd_base_module(stream, disttag_marking, error_msg, require_empty_database):
    mmd = load_mmd(read_staged_data("platform"))
    mmd = mmd.copy(mmd.get_module_name(), stream)

    if disttag_marking:
        xmd = mmd.get_xmd()
        xmd["mbs"]["disttag_marking"] = disttag_marking
        mmd.set_xmd(xmd)

    if error_msg:
        with pytest.raises(UnprocessableEntity, match=error_msg):
            import_mmd(db_session, mmd)
    else:
        import_mmd(db_session, mmd)


def test_import_mmd_remove_dropped_virtual_streams():
    mmd = load_mmd(read_staged_data("formatted_testmodule"))

    # Add some virtual streams
    xmd = mmd.get_xmd()
    xmd["mbs"]["virtual_streams"] = ["f28", "f29", "f30"]
    mmd.set_xmd(xmd)

    # Import mmd into database to simulate the next step to reimport a module
    import_mmd(db_session, mmd)

    # Now, remove some virtual streams from module metadata
    xmd = mmd.get_xmd()
    xmd["mbs"]["virtual_streams"] = ["f28", "f29"]  # Note that, f30 is removed
    mmd.set_xmd(xmd)

    # Test import modulemd again and the f30 should be removed from database.
    module_build, _ = import_mmd(db_session, mmd)

    db_session.refresh(module_build)
    assert ["f28", "f29"] == sorted(item.name for item in module_build.virtual_streams)
    assert 0 == db_session.query(models.VirtualStream).filter_by(name="f30").count()


def test_import_mmd_dont_remove_dropped_virtual_streams_associated_with_other_modules():
    mmd = load_mmd(read_staged_data("formatted_testmodule"))
    # Add some virtual streams to this module metadata
    xmd = mmd.get_xmd()
    xmd["mbs"]["virtual_streams"] = ["f28", "f29", "f30"]
    mmd.set_xmd(xmd)
    import_mmd(db_session, mmd)

    # Import another module which has overlapping virtual streams
    another_mmd = load_mmd(read_staged_data("formatted_testmodule-more-components"))
    # Add some virtual streams to this module metadata
    xmd = another_mmd.get_xmd()
    xmd["mbs"]["virtual_streams"] = ["f29", "f30"]
    another_mmd.set_xmd(xmd)
    another_module_build, _ = import_mmd(
        db_session, another_mmd)

    # Now, remove f30 from mmd
    xmd = mmd.get_xmd()
    xmd["mbs"]["virtual_streams"] = ["f28", "f29"]
    mmd.set_xmd(xmd)

    # Reimport formatted_testmodule again
    module_build, _ = import_mmd(db_session, mmd)

    db_session.refresh(module_build)
    assert ["f28", "f29"] == sorted(item.name for item in module_build.virtual_streams)

    # The overlapped f30 should be still there.
    db_session.refresh(another_module_build)
    assert ["f29", "f30"] == sorted(item.name for item in another_module_build.virtual_streams)



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT
from __future__ import absolute_import

import mock

import module_build_service.common.models
from module_build_service.common.models import ComponentBuild
from module_build_service.scheduler.db_session import db_session
import module_build_service.scheduler.handlers.repos
from tests import scheduler_init_data


class TestRepoDone:

    @mock.patch("module_build_service.common.models.ModuleBuild.get_by_tag")
    def test_no_match(self, get_by_tag):
        """ Test that when a repo msg hits us and we have no match,
        that we do nothing gracefully.
        """
        scheduler_init_data()
        get_by_tag.return_value = None
        module_build_service.scheduler.handlers.repos.done(
            msg_id="no matches for this...",
            tag_name="2016-some-nonexistent-build")

    @mock.patch(
        "module_build_service.builder.KojiModuleBuilder."
        "KojiModuleBuilder.recover_orphaned_artifact",
        return_value=[],
    )
    @mock.patch(
        "module_build_service.builder.KojiModuleBuilder."
        "KojiModuleBuilder.get_average_build_time",
        return_value=0.0,
    )
    @mock.patch(
        "module_build_service.builder.KojiModuleBuilder."
        "KojiModuleBuilder.list_tasks_for_components",
        return_value=[],
    )
    @mock.patch(
        "module_build_service.builder.KojiModuleBuilder.KojiModuleBuilder.buildroot_ready",
        return_value=True,
    )
    @mock.patch("module_build_service.builder.KojiModuleBuilder.get_session")
    @mock.patch("module_build_service.builder.KojiModuleBuilder.KojiModuleBuilder.build")
    @mock.patch(
        "module_build_service.builder.KojiModuleBuilder.KojiModuleBuilder.buildroot_connect"
    )
    def test_a_single_match(
        self, connect, build_fn, get_session, ready, list_tasks_fn, mock_gabt, mock_uea
    ):
        """ Test that when a repo msg hits us and we have a single match.
        """
        scheduler_init_data()
        get_session.return_value = mock.Mock(), "development"
        build_fn.return_value = 1234, 1, "", None

        module_build_service.scheduler.handlers.repos.done(
            msg_id="some_msg_id",
            tag_name="module-testmodule-master-20170109091357-7c29193d-build")
        build_fn.assert_called_once_with(
            artifact_name="tangerine",
            source=(
                "https://src.fedoraproject.org/rpms/tangerine?"
                "#fbed359411a1baa08d4a88e0d12d426fbf8f602c"
            ),
        )

    @mock.patch("module_build_service.builder.KojiModuleBuilder.KojiModuleBuilder.finalize")
    @mock.patch(
        "module_build_service.builder.KojiModuleBuilder."
        "KojiModuleBuilder.recover_orphaned_artifact",
        return_value=[],
    )
    @mock.patch(
        "module_build_service.builder.KojiModuleBuilder."
        "KojiModuleBuilder.get_average_build_time",
        return_value=0.0,
    )
    @mock.patch(
        "module_build_service.builder.KojiModuleBuilder."
        "KojiModuleBuilder.list_tasks_for_components",
        return_value=[],
    )
    @mock.patch(
        "module_build_service.builder.KojiModuleBuilder.KojiModuleBuilder.buildroot_ready",
        return_value=True,
    )
    @mock.patch("module_build_service.builder.KojiModuleBuilder.get_session")
    @mock.patch("module_build_service.builder.KojiModuleBuilder.KojiModuleBuilder.build")
    @mock.patch(
        "module_build_service.builder.KojiModuleBuilder.KojiModuleBuilder.buildroot_connect"
    )
    def test_a_single_match_finalize(
        self, connect, build_fn, get_session, ready, list_tasks_fn, mock_gabt, mock_uea, finalizer
    ):
        """ Test that when a repo msg hits us and we have a single match.
        """
        scheduler_init_data(tangerine_state=1)
        get_session.return_value = mock.Mock(), "development"
        build_fn.return_value = 1234, 1, "", None

        # Ensure the time_completed is None, so we can test it is set to
        # some date once the build is finalized.
        module_build = module_build_service.common.models.ModuleBuild.get_by_id(db_session, 2)
        module_build.time_completed = None
        db_session.commit()

        def mocked_finalizer(succeeded=None):
            # Check that the time_completed is set in the time when
            # finalizer is called.
            assert succeeded is True
            module_build = module_build_service.common.models.ModuleBuild.get_by_id(db_session, 2)
            assert module_build.time_completed is not None

        finalizer.side_effect = mocked_finalizer

        module_build_service.scheduler.handlers.repos.done(
            msg_id="some_msg_id",
            tag_name="module-testmodule-master-20170109091357-7c29193d-build")

        finalizer.assert_called_once()

    @mock.patch(
        "module_build_service.builder.KojiModuleBuilder."
        "KojiModuleBuilder.recover_orphaned_artifact",
        return_value=[],
    )
    @mock.patch(
        "module_build_service.builder.KojiModuleBuilder."
        "KojiModuleBuilder.get_average_build_time",
        return_value=0.0,
    )
    @mock.patch(
        "module_build_service.builder.KojiModuleBuilder."
        "KojiModuleBuilder.list_tasks_for_components",
        return_value=[],
    )
    @mock.patch(
        "module_build_service.builder.KojiModuleBuilder.KojiModuleBuilder.buildroot_ready",
        return_value=True,
    )
    @mock.patch("module_build_service.builder.KojiModuleBuilder.get_session")
    @mock.patch("module_build_service.builder.KojiModuleBuilder.KojiModuleBuilder.build")
    @mock.patch(
        "module_build_service.builder.KojiModuleBuilder.KojiModuleBuilder.buildroot_connect"
    )
    def test_a_single_match_build_fail(
        self, connect, build_fn, config, ready, list_tasks_fn, mock_gabt, mock_uea
    ):
        """ Test that when a KojiModuleBuilder.build fails, the build is
        marked as failed with proper state_reason.
        """
        scheduler_init_data()
        config.return_value = mock.Mock(), "development"
        build_fn.return_value = None, 4, "Failed to submit artifact tangerine to Koji", None

        module_build_service.scheduler.handlers.repos.done(
            msg_id="some_msg_id",
            tag_name="module-testmodule-master-20170109091357-7c29193d-build")

        build_fn.assert_called_once_with(
            artifact_name="tangerine",
            source=(
                "https://src.fedoraproject.org/rpms/tangerine?"
                "#fbed359411a1baa08d4a88e0d12d426fbf8f602c"
            ),
        )
        component_build = db_session.query(
            module_build_service.common.models.ComponentBuild
        ).filter_by(package="tangerine").one()
        assert component_build.state_reason == "Failed to submit artifact tangerine to Koji"

    @mock.patch("module_build_service.scheduler.handlers.repos.log.info")
    def test_erroneous_regen_repo_received(self, mock_log_info):
        """ Test that when an unexpected KojiRepoRegen message is received, the module doesn't
        complete or go to the next build batch.
        """
        scheduler_init_data(1)

        component_build = db_session.query(ComponentBuild).filter_by(package="tangerine").one()
        component_build.tagged = False
        db_session.commit()

        module_build_service.scheduler.handlers.repos.done(
            msg_id="some_msg_id",
            tag_name="module-testmodule-master-20170109091357-7c29193d-build")

        mock_log_info.assert_called_with(
            "Ignoring repo regen, because not all components are tagged."
        )
        module_build = module_build_service.common.models.ModuleBuild.get_by_id(db_session, 2)
        # Make sure the module build didn't transition since all the components weren't tagged
        assert module_build.state == module_build_service.common.models.BUILD_STATES["build"]

    @mock.patch(
        "module_build_service.builder.KojiModuleBuilder."
        "KojiModuleBuilder.list_tasks_for_components",
        return_value=[],
    )
    @mock.patch(
        "module_build_service.builder.KojiModuleBuilder.KojiModuleBuilder.buildroot_ready",
        return_value=True,
    )
    @mock.patch("module_build_service.builder.KojiModuleBuilder.get_session")
    @mock.patch("module_build_service.builder.KojiModuleBuilder.KojiModuleBuilder.build")
    @mock.patch(
        "module_build_service.builder.KojiModuleBuilder.KojiModuleBuilder.buildroot_connect"
    )
    @mock.patch(
        "module_build_service.builder.GenericBuilder.default_buildroot_groups",
        return_value={"build": [], "srpm-build": []},
    )
    def test_failed_component_build(
        self, dbg, connect, build_fn, config, ready, list_tasks_fn
    ):
        """ Test that when a KojiModuleBuilder.build fails, the build is
        marked as failed with proper state_reason.
        """
        scheduler_init_data(3)
        config.return_value = mock.Mock(), "development"
        build_fn.return_value = None, 4, "Failed to submit artifact x to Koji", None

        module_build_service.scheduler.handlers.repos.done(
            msg_id="some_msg_id",
            tag_name="module-testmodule-master-20170109091357-7c29193d-build")

        module_build = module_build_service.common.models.ModuleBuild.get_by_id(db_session, 2)
        assert module_build.state == module_build_service.common.models.BUILD_STATES["failed"]



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT
from __future__ import absolute_import

import mock

from module_build_service.common.koji import get_session


@mock.patch("koji.ClientSession")
def test_get_anonymous_session(mock_session):
    mbs_config = mock.Mock(koji_profile="koji", koji_config="conf/koji.conf")
    session = get_session(mbs_config, login=False)
    assert mock_session.return_value == session
    assert mock_session.return_value.krb_login.assert_not_called
